from bppy.model.event_selection.smt_event_selection_strategy import SMTEventSelectionStrategy

def ExperimentalSMTEventSelectionStrategy():
    from warnings import warn
    warn('Class ExperimentalSMTEventSelectionStrategy is deprecated. Returned SMTEventSelectionStrategy instead.',
         DeprecationWarning, stacklevel=2)
    return SMTEventSelectionStrategy()



from bppy import *

hot = Bool('hot')
cold = Bool('cold')

@b_thread
def three_hot():
    for i in range(3):
        yield {request: hot , waitFor: hot}

@b_thread
def three_cold():
    for j in range(3):
        yield {request: cold , waitFor: cold}

@b_thread
def control_temp():
    m = yield {}
    while True:
        if is_true(m[cold]):
            m = yield {block: cold}
        if is_true(m[hot]):
            m = yield {block: hot}

@b_thread
def mutual_exclusion():
    yield {block: And(cold ,hot), waitFor: false}


if __name__ == "__main__":
    b_program = BProgram(bthreads=[three_cold(), three_hot(), mutual_exclusion(), control_temp()],
                         event_selection_strategy=SMTEventSelectionStrategy(),
                         listener=PrintBProgramRunnerListener())
    b_program.run()



from bppy import *


@b_thread
def exponential_growth(exp):
    yield {request: BEvent(f"{exp}")}
    if exp > 0:
        b_program.add_bthread(exponential_growth(exp - 1))
        b_program.add_bthread(exponential_growth(exp - 1))


if __name__ == "__main__":
    b_program = BProgram(bthreads=[exponential_growth(4)],
                         event_selection_strategy=SimpleEventSelectionStrategy(),
                         listener=PrintBProgramRunnerListener())
    b_program.run()



from bppy import *
# import pygame

H = 250
W = 250

N = 5

p = Real('p')

X = RealVector('x', N)
Y = RealVector('y', N)

dX = RealVector('dx', N)
dY = RealVector('dy', N)

dT = 0.05
speed = 0.2

nextX = [X[i] + dX[i] * dT for i in range(N)]
nextY = [Y[i] + dY[i] * dT for i in range(N)]


def center_of_mass():
    req = Or(And(sum(dX) == speed, sum(dY) == speed, p == 3),
             And(sum(dX) == 0, sum(dY) == -speed, p == 2))

    yield {'request': req, 'wait-for': false}


def obstacle(i):
    obs = And(nextX[i] >= 0.7, nextX[i] <= 0.75,
              nextY[i] >= 0.7, nextY[i] <= 0.75)

    yield {'block': obs, 'wait-for': false}


def structure_x():
    yield {'block': Or([Not(dX[i] == dX[i - 1]) for i in range(1, N)]), 'wait-for': false}


def structure_y():
    yield {'block': Or([Not(dY[i] == dY[i - 1]) for i in range(1, N)]), 'wait-for': false}


def init_x():
    yield {'request': And([X[i] == 0.1 * i for i in range(N)])}


def init_y():
    yield {'request': And([Y[i] == 0.1 * i for i in range(N)])}


def step(i):
    m = yield {}
    while True:
        d = And(X[i] == m.eval(nextX[i]), Y[i] == m.eval(nextY[i]))
        m = yield {'request': d, 'block': Not(d)}


def walls(i):
    yield {'block': Or(nextX[i] < 0, nextX[i] > 1, nextY[i] < 0, nextY[i] > 1), 'wait-for': false}

# TODO: debug display function - issue with pygame
# def display():
#     pygame.init()
#     screen = pygame.display.set_mode((W + 15, H + 15))
#
#     clock = pygame.time.Clock()
#
#     while True:
#         for event in pygame.event.get():
#             if event.type == pygame.QUIT:
#                 yield {'block': true for i in range(N)}
#
#         m = yield {}
#         screen.fill((250, 250, 250))
#
#         for i in range(N):
#             x = toFloat(m[X[i]])
#             y = toFloat(m[Y[i]])
#             pygame.draw.rect(screen, (50, 100, 0),
#                              pygame.Rect(5 + x * H, 5 + y * W, 5, 5))
#
#         pygame.draw.rect(screen, (0, 50, 255),
#                          pygame.Rect(11.25 + 0.7 * H, 11.25 + 0.7 * W, 10, 10))
#
#         pygame.display.flip()
#         clock.tick(60)


if __name__ == "__main__":
    b_program = BProgram(

        bthreads=[init_x(), init_y(), center_of_mass(), structure_x(),
                  structure_y()] +
                 [walls(i) for i in range(N)] +
                 [step(i) for i in range(N)] +
                 [obstacle(i) for i in range(N)],

        event_selection_strategy=SMTEventSelectionStrategy(),
        listener=PrintBProgramRunnerListener())
    b_program.event_selection_strategy.debug = True

    b_program.run()



from bppy.execution.listeners.b_program_runner_listener import BProgramRunnerListener


class PrintBProgramRunnerListener(BProgramRunnerListener):
    """
    This class implements the :class:`BProgramRunnerListener
    <bppy.execution.listeners.b_program_runner_listener.BProgramRunnerListener>` interface, providing basic print
    statements for specific events in the BProgram's execution.
    """
    def starting(self, b_program):
        """
        Prints "STARTED" when the BProgram execution is about to start.
        """
        print("STARTED")

    def started(self, b_program):
        pass

    def super_step_done(self, b_program):
        pass

    def ended(self, b_program):
        """
        Prints "ENDED" when the BProgram execution is about to start.
        """
        print("ENDED")

    def assertion_failed(self, b_program):
        pass

    def b_thread_added(self, b_program):
        pass

    def b_thread_removed(self, b_program):
        pass

    def b_thread_done(self, b_program):
        pass

    def event_selected(self, b_program, event):
        """
        Prints the selected event when an event has been selected during the BProgram's execution.
        """
        print(event)

    def halted(self, b_program):
        pass

    def __init__(self):
        """
        Initializes the PrintBProgramRunnerListener.
        """
        super().__init__()





from bppy.model.event_selection.event_selection_strategy import EventSelectionStrategy
from abc import abstractmethod

class SolverBasedEventSelectionStrategy(EventSelectionStrategy):
    """
    A base class used to represent a Solver based Event Selection Strategy. This is an abstract class
    that requires the implementation of `select` and `is_satisfied` methods.
    """
    @abstractmethod
    def select(self, statements, external_events_queue=[]):
        """
        Abstract method to select and return the bprogram's next event. For more information,
        see :class:`SMTEventSelectionStrategy
        <bppy.model.event_selection.smt_event_selection_strategy.SMTEventSelectionStrategy>`.
        """
        pass

    @abstractmethod
    def is_satisfied(self, event, statement):
        """
        Abstract method to check whether a given event satisfies the given sync statement, and the bthread should
        advance. For more information, see :class:`SMTEventSelectionStrategy
        <bppy.model.event_selection.smt_event_selection_strategy.SMTEventSelectionStrategy>`.
        """
        pass




from copy import copy


def b_thread(func):
    """
    A decorator to wrap bthread generator with, in order to handle data transmission and bthread termination.
    """
    def wrapper(*args):
        while True:
            m = None
            f = func(*args)
            while True:
                try:
                    e = f.send(m)
                    m = yield e
                    if m is None:
                        break
                except (KeyError, StopIteration):
                    m = yield None
                    break
    return wrapper




from bppy.model.b_event import BEvent
from inspect import signature


class EventSet:
    """
    A class to represent a symbolically defined set of events using a given predicate.

    Attributes
    ----------
    predicate : callable
        The predicate that tests if an event is a member of this event set.
    data : dict
        Additional keyword arguments that could be used by the predicate.
    """

    def __init__(self, predicate, **kwargs):
        """
        Constructs all the necessary attributes for the EventSet object.

        Parameters
        ----------
        predicate : callable
            The predicate that tests if an event is a member of this event set. It should be a function
            that takes an event as its first argument and returns a boolean.
        **kwargs : dict
            Additional keyword arguments that could be used by the predicate.
        """
        self.predicate = predicate
        self.data = kwargs

    def __contains__(self, event):
        if len(signature(self.predicate).parameters) > 1:
            return self.predicate(event, self.data)
        else:
            return self.predicate(event)


class All(EventSet):
    """
    A class to represent the set of all events.
    """
    def __init__(self):
        super().__init__(lambda e: True)


class EmptyEventSet(EventSet):
    """
    A class to represent an empty event set.
    """
    def __init__(self):
        super().__init__(lambda e: False)


class AllExcept(EventSet):
    """
    A class to represent a set of all events except some specified event or event set.
    """
    def __init__(self, event):
        """
        Constructs all the necessary attributes for the AllExcept object.

        If the input event is an instance of BEvent, an EventSet is created with a predicate that
        matches all events except for the input event. If the input event is an EventSet, an EventSet is
        created with a predicate that matches all events not contained in the input EventSet.

        Parameters
        ----------
        event : `BEvent <bppy.model.b_event.BEvent>` or `EventSet <bppy.model.event_set.EventSet>`
            The event or set of events to be excluded from the event set.
        """
        if isinstance(event, BEvent):
            super().__init__(lambda e: event != e)
        else:  # Event Set
            super().__init__(lambda e: not event.__contains__(e))


class EventSetList(EventSet):
    """
    A class to represent a union of event sets and events.
    """
    def __init__(self, lst):
        """
        Constructs all the necessary attributes for the EventSetList object.

        Parameters
        ----------
        lst : list
            A list of `BEvent <bppy.model.b_event.BEvent>` or `EventSet <bppy.model.event_set.EventSet>` objects.
            The resulting EventSetList will contain any event that is equal to a BEvent
            in the list or is contained in an EventSet in the list.
        """
        self.lst = lst
        super().__init__(lambda e: any([EventSetList._item_contains(item, e) for item in self.lst]))

    @staticmethod
    def _item_contains(item, event):
        if isinstance(item, BEvent):
            return event == item
        elif len(signature(item.predicate).parameters) > 1:
            return item.predicate(event, item.data)
        else:
            return item.predicate(event)




from bppy import *


@b_thread
def add_hot():
    yield {request: BEvent("HOT")}
    yield {request: BEvent("HOT")}
    yield {request: BEvent("HOT")}


@b_thread
def add_cold():
    yield {request: BEvent("COLD")}
    yield {request: BEvent("COLD")}
    yield {request: BEvent("COLD")}


@b_thread
def control_temp():
    while True:
        yield {waitFor: BEvent("COLD"), block: BEvent("HOT")}
        yield {waitFor: BEvent("HOT"), block: BEvent("COLD")}


if __name__ == "__main__":
    b_program = BProgram(bthreads=[add_hot(), add_cold(), control_temp()],
                         event_selection_strategy=SimpleEventSelectionStrategy(),
                         listener=PrintBProgramRunnerListener())
    b_program.run()



from bppy.model.event_selection.simple_event_selection_strategy import SimpleEventSelectionStrategy
from collections.abc import Iterable
from bppy.model.b_event import BEvent


class PriorityBasedEventSelectionStrategy(SimpleEventSelectionStrategy):
    """
    Priority-based extension for :class:`SimpleEventSelectionStrategy
    <bppy.model.event_selection.event_selection_strategy.SimpleEventSelectionStrategy>`, selecting the requested
    non-blocked event with the highest priority. If no priority is specified for an event, the `default_priority`
    attribute is used.

    Attributes
    ----------
    default_priority : int
        The default priority value to use if no priority is specified for an event.
    """
    def __init__(self, default_priority=0):
        """
        Constructs a new PriorityBasedEventSelectionStrategy with a given default priority.

        Parameters
        ----------
        default_priority : int, optional
            The default priority to be used when no specific priority is given for an event.
        """
        self.default_priority = default_priority

    def selectable_events(self, statements):
        """
        Returns a set of possible events that can be selected from a list of bthread statements. Specifically,
        the method extracts requested non-blocked events, with the maximal priority value, from the provided bthreads statements.

        Parameters
        ----------
        statements : list
            A list of bthread statements from which to extract possible events.

        Returns
        -------
        set
            A set of possible events that can be selected.
        """
        possible_events = set()
        for statement in statements:
            if 'request' in statement:  # should be eligible for sets
                p = statement.get('priority', self.default_priority)
                if isinstance(statement['request'], Iterable):
                    possible_events.update([(x, p) for x in statement['request']])
                elif isinstance(statement['request'], BEvent):
                    possible_events.add((statement['request'], p))
                else:
                    raise TypeError("request parameter should be BEvent or iterable")
        for statement in statements:
            if 'block' in statement:
                if isinstance(statement.get('block'), BEvent):
                    possible_events = {(x, p) for x, p in possible_events if x != statement.get('block')}
                else:
                    possible_events = {(x, p) for x, p in possible_events if x not in statement.get('block')}
        if len(possible_events) == 0:
            return possible_events
        max_priority = max([p for _, p in possible_events])
        return {x for x, p in possible_events if p == max_priority}



from bppy import *

x = lambda row, col: BEvent('X' + str(row) + str(col))
o = lambda row, col: BEvent('O' + str(row) + str(col))

LINES = [[(i, j) for j in range(3)] for i in range(3)] + [[(i, j) for i in range(3)] for j in range(3)] + [
    [(i, i) for i in range(3)]] + [[(i, 3 - i - 1) for i in range(3)]]
x_lines = [[x(i, j) for (i, j) in line] for line in LINES]
o_lines = [[o(i, j) for (i, j) in line] for line in LINES]

any_x = [x(i, j) for i in range(3) for j in range(3)]
any_o = [o(i, j) for i in range(3) for j in range(3)]
move_events = EventSet(lambda e: e.name.startswith('X') or e.name.startswith('O'))

static_event = {
    'OWin': BEvent('OWin'),
    'XWin': BEvent('XWin'),
    'draw': BEvent('Draw')
}


@b_thread
def square_taken(row, col):
    yield {waitFor: [x(row, col), o(row, col)]}
    yield {block: [x(row, col), o(row, col)]}


@b_thread
def enforce_turns():
    while True:
        yield {waitFor: any_x, block: any_o}
        yield {waitFor: any_o, block: any_x}


@b_thread
def end_of_game():
    yield {waitFor: list(static_event.values())}
    yield {block: All()}


@b_thread
def detect_draw():
    for r in range(3):
        for c in range(3):
            yield {waitFor: move_events}
    yield {request: static_event['draw'], priority: 90}


@b_thread
def detect_x_win(line):
    for i in range(3):
        yield {waitFor: line}
    yield {request: static_event['XWin'], priority: 100}


@b_thread
def detect_o_win(line):
    for i in range(3):
        yield {waitFor: line}
    yield {request: static_event['OWin'], priority: 100}


# Preference to put O on the center
@b_thread
def center_preference():
    while True:
        yield {request: o(1, 1), priority: 35}


# Preference to put O on the corners
@b_thread
def corner_preference():
    while True:
        yield {request: [o(0, 0), o(0, 2), o(2, 0), o(2, 2)], priority: 20}


# Preference to put O on the sides
@b_thread
def side_preference():
    while True:
        yield {request: [o(0, 1), o(1, 0), o(1, 2), o(2, 1)], priority: 10}


# player O strategy to add a third O to win
@b_thread
def add_third_o(line):
    for i in range(2):
        yield {waitFor: line}
    yield {request: line, priority: 50}


# player O strategy to prevent a third X
@b_thread
def prevent_third_x(xline, oline):
    for i in range(2):
        yield {waitFor: xline}
    yield {request: oline, priority: 40}


@b_thread
def block_fork(xfork, ofork):
    for i in range(2):
        yield {waitFor: xfork}
    yield {request: ofork, priority: 30}


forks22 = [[x(1, 2), x(2, 0)], [x(2, 1), x(0, 2)], [x(1, 2), x(2, 1)]], [o(2, 2), o(0, 2), o(2, 0)]
forks02 = [[x(1, 2), x(0, 0)], [x(0, 1), x(2, 2)], [x(1, 2), x(0, 1)]], [o(0, 2), o(0, 0), o(2, 2)]
forks20 = [[x(1, 0), x(2, 2)], [x(2, 1), x(0, 0)], [x(2, 1), x(1, 0)]], [o(2, 0), o(0, 0), o(2, 2)]
forks00 = [[x(0, 1), x(2, 0)], [x(1, 0), x(0, 2)], [x(0, 1), x(1, 0)]], [o(0, 0), o(0, 2), o(2, 0)]
forks_diag = [[x(0, 2), x(2, 0)], [x(0, 0), x(2, 2)]], [o(0, 1), o(1, 0), o(1, 2), o(2, 1)]



# simulate player X
@b_thread
def player_x():
    while True:
        yield {request: any_x}


if __name__ == "__main__":
    bprog = BProgram(
        bthreads=[square_taken(i, j) for i in range(3) for j in range(3)] +
                 [enforce_turns(), end_of_game(), detect_draw()] +
                 [detect_x_win(line) for line in x_lines] +
                 [detect_o_win(line) for line in o_lines] +
                 [center_preference(), corner_preference(), side_preference()] +
                 [add_third_o(line) for line in o_lines] +
                 [prevent_third_x(xline, oline) for (xline, oline) in zip(x_lines, o_lines)] +
                 [block_fork(xfork, forks22[1]) for xfork in forks22[0]] +
                 [block_fork(xfork, forks02[1]) for xfork in forks02[0]] +
                 [block_fork(xfork, forks20[1]) for xfork in forks20[0]] +
                 [block_fork(xfork, forks00[1]) for xfork in forks00[0]] +
                 [block_fork(xfork, forks_diag[1]) for xfork in forks_diag[0]] +
                 [player_x()],
        event_selection_strategy=PriorityBasedEventSelectionStrategy(default_priority=0),
        listener=PrintBProgramRunnerListener()
    )
    bprog.run()



from abc import ABC, abstractmethod


class BProgramRunnerListener(ABC):
    """
    Abstract Base Class representing a listener for :class:`BProgram <bppy.model.bprogram.BProgram>` execution events.
    """
    @abstractmethod
    def starting(self, b_program):
        """
        Abstract method to handle the start of the execution of a :class:`BProgram <bppy.model.bprogram.BProgram>`.
        """
        pass

    @abstractmethod
    def started(self, b_program):
        """
        Abstract method to handle the start of the execution of a :class:`BProgram <bppy.model.bprogram.BProgram>`.
        """
        pass

    @abstractmethod
    def super_step_done(self, b_program):
        """
        Abstract method to handle the completion of a super-step of a :class:`BProgram <bppy.model.bprogram.BProgram>`.
        """
        pass

    @abstractmethod
    def ended(self, b_program):
        """
        Abstract method to handle the end of the execution of a :class:`BProgram <bppy.model.bprogram.BProgram>`.
        """
        pass

    @abstractmethod
    def assertion_failed(self, b_program):
        """
        Abstract method to handle assertion failures in the execution of a :class:`BProgram <bppy.model.bprogram.BProgram>`.
        """
        pass

    @abstractmethod
    def b_thread_added(self, b_program):
        """
        Abstract method to handle the addition of a bthread in the :class:`BProgram <bppy.model.bprogram.BProgram>`.
        """
        pass

    @abstractmethod
    def b_thread_removed(self, b_program):
        """
        Abstract method to handle the removal of a bthread in the :class:`BProgram <bppy.model.bprogram.BProgram>`.
        """
        pass

    @abstractmethod
    def b_thread_done(self, b_program):
        """
        Abstract method to handle the termination of a bthread in the execution of a :class:`BProgram <bppy.model.bprogram.BProgram>`.
        """
        pass

    @abstractmethod
    def event_selected(self, b_program, event):
        """
        Abstract method to handle the selection of event during the execution of a :class:`BProgram <bppy.model.bprogram.BProgram>`.
        """
        pass

    @abstractmethod
    def halted(self, b_program):
        """
        Abstract method to handle the halting of the execution of a :class:`BProgram <bppy.model.bprogram.BProgram>`.
        """
        pass




from importlib import import_module
from inspect import getmembers, isfunction
from itertools import tee
import warnings

from z3 import *


class BProgram:
    """
    A class to represent a Behavioral Program (BProgram) object.

    Attributes
    ----------
    source_name : str
        A file name from which bthreads are imported.
    new_bt : list
        A list of new bthreads that have not been loaded yet.
    bthreads : list
        A list of bthreads to be run in the BProgram.
    event_selection_strategy : :class:`EventSelectionStrategy <bppy.model.event_selection.event_selection_strategy.EventSelectionStrategy>`
        A strategy object to select events and advance bthreads.
    listener : :class:`BProgramRunnerListener <bppy.execution.listeners.b_program_runner_listener.BProgramRunnerListener>`
        A listener object to be invoked at the start and end of the program, and at each event selection.
    variables : dict
        A dictionary of variables declared in the source module (for SMT based event selection startegy).
    tickets : list
        A list of bthread tickets (used in execution).
    external_events_queue : list
        A queue to handle external events introduced into the BProgram.
    """
    def __init__(self, bthreads=None, source_name=None, event_selection_strategy=None, listener=None):
        """
        Constructs all the necessary attributes for the BProgram object.

        Parameters
        ----------
        bthreads : list, optional
            A list of bthreads to be run in the BProgram.
        source_name : str, optional
            The module from which bthreads are imported. If this is provided,
            the bthreads and variables from the source module will be imported when setup() is called.
        event_selection_strategy : :class:`EventSelectionStrategy <bppy.model.event_selection.event_selection_strategy.EventSelectionStrategy>`, optional
            An event selection strategy object.
        listener : :class:`BProgramRunnerListener <bppy.execution.listeners.b_program_runner_listener.BProgramRunnerListener>`, optional
            A bprogram listener object.
        """
        self.source_name = source_name
        self.new_bt = []
        self.bthreads = bthreads
        self.event_selection_strategy = event_selection_strategy
        self.listener = listener
        self.variables = None
        self.tickets = []
        self.external_events_queue = []

    def setup(self):
        """
        Sets up the BProgram.

        If a source file is provided, this method imports all bthreads and variables
        into the BProgram. All bthreads are then loaded.
        """
        if self.source_name:
            self.bthreads = [o[1]() for o in getmembers(import_module(self.source_name)) if
                             isfunction(o[1]) and o[1].__module__ == self.source_name]

            self.variables = dict([o for o in getmembers(import_module(self.source_name)) if
                                   isinstance(o[1], ExprRef) or isinstance(o[1], list)])

        self.new_bt = self.bthreads
        self.load_new_bthreads()


    def advance_bthreads(self,tickets, m):
        """
        Advances bthreads based on the selected event (m), using the event selection strategy satisfactions criteria (e.g., event was requested or waited for).

        Parameters
        ----------
        tickets : list
            A list of bthread tickets.
        m : :class:`BEvent <bppy.model.b_event.BEvent>`
            The selected event to be sent as message to the bthreads.
        """
        if len(self.new_bt) > 0:
            warnings.warn("Some new bthreads are not loaded and are not affecting the bprogram. Use the load_new_bthreads method to load them.", RuntimeWarning)
        for l in tickets:
            if m is None or self.event_selection_strategy.is_satisfied(m, l):
                try:
                    bt = l['bt']
                    l.clear()
                    ll = bt.send(m)
                    if ll is None:
                        continue
                    l.update(ll)
                    l.update({'bt': bt})
                except (KeyError, StopIteration):
                    pass

    def add_bthread(self, bt):
        """
        Adds a new bthread to the BProgram (should be loaded after using :func:`load_new_bthreads <bppy.model.bprogram.BProgram.load_new_bthreads>`).

        Parameters
        ----------
        bt : Any
            The bthread to be added.
        """
        self.new_bt.append(bt)

    def load_new_bthreads(self):
        """
        Loads all new bthreads into the BProgram.

        The method uses :func:`advance_bthreads <bppy.model.bprogram.BProgram.advance_bthreads>` to load bthreads.
        """
        while len(self.new_bt) > 0:
            new_tickets = [{'bt': bt} for bt in self.new_bt]
            self.new_bt.clear()
            self.advance_bthreads(new_tickets, None)
            self.tickets.extend(new_tickets)

    def next_event(self):
        """
        Selects the next event in the BProgram using the event selection strategy.

        If there are new bthreads that haven't been loaded, a warning will be raised.

        Returns
        -------
        event : :class:`BEvent <bppy.model.b_event.BEvent>`
            The selected event to be processed.
        """
        if len(self.new_bt) > 0:
            warnings.warn(
                "Some new bthreads are not loaded and are not affecting the bprogram. Use the load_new_bthreads method to load them.",
                RuntimeWarning)
        return self.event_selection_strategy.select(self.tickets, self.external_events_queue)

    def run(self):
        """
        Runs the BProgram.

        It sets up the BProgram, selects events, and advance the bthreads until no event can be selected
        or until interrupted by the listener. At the start and end of the run, and after each
        event selection, the listener (if provided) is invoked.
        """
        if self.listener:
            self.listener.starting(b_program=self)

        self.setup()
        # Main loop
        interrupted = False
        while not interrupted:
            # for dynamically added bthreads
            self.load_new_bthreads()

            event = self.next_event()
            # Finish the program if no event is selected
            if event is None:
                break
            if self.listener:
                interrupted = self.listener.event_selected(b_program=self, event=event)

            self.advance_bthreads(self.tickets,event)


        if self.listener:
            self.listener.ended(b_program=self)

    def enqueue_external_event(self, event):
        """
        Enqueues an external event into the BProgram.

        Parameters
        ----------
        event : :class:`BEvent <bppy.model.b_event.BEvent>`
            The external event to be enqueued.
        """
        self.external_events_queue.append(event)



from bppy import *


@b_thread
def hello():
    yield {request: BEvent("Hello")}

@b_thread
def world():
    yield {request: BEvent("World")}


if __name__ == "__main__":
    b_program = BProgram(bthreads=[hello(), world()],
                         event_selection_strategy=SimpleEventSelectionStrategy(),
                         listener=PrintBProgramRunnerListener())
    b_program.run()



from bppy.model.event_selection.event_selection_strategy import EventSelectionStrategy
from bppy.model.b_event import BEvent
from bppy.model.event_set import EmptyEventSet
import random
from collections.abc import Iterable


class SimpleEventSelectionStrategy(EventSelectionStrategy):
    """
    A simple :class:`EventSelectionStrategy
    <bppy.model.event_selection.event_selection_strategy.EventSelectionStrategy>`, which uniformly selects an event
    that is requested and blocked. It advances all bthreads that requested or waited for the selected event.
    """
    def is_satisfied(self, event, statement):
        """
        Checks whether a bthread should advance based on the selected event and its current sync statement.
        Specifically, It checks whether the statement requests or waits for the selected event.

        Parameters
        ----------
        event : :class:`BEvent <bppy.model.b_event.BEvent>`
            The selected event to be checked against the statement.
        statement : dict
            The bthread's sync statement to check the selected event against.

        Returns
        -------
        bool
            True if the statement requests or waits for the selected event, False otherwise.
        """
        if isinstance(statement.get('request'), BEvent):
            if isinstance(statement.get('waitFor'), BEvent):
                return statement.get('request') == event or statement.get('waitFor') == event
            else:
                return statement.get('request') == event or statement.get('waitFor', EmptyEventSet()).__contains__(event)
        else:
            if isinstance(statement.get('waitFor'), BEvent):
                return statement.get('request', EmptyEventSet()).__contains__(event) or statement.get('waitFor') == event
            else:
                return statement.get('request', EmptyEventSet()).__contains__(event) or statement.get('waitFor', EmptyEventSet()).__contains__(event)

    def selectable_events(self, statements):
        """
        Returns a set of possible events that can be selected from a list of bthread statements. Specifically,
        the method extracts requested non-blocked events from the provided bthreads statements.

        Parameters
        ----------
        statements : list
            A list of bthreads sync statements from which to extract selectable events.

        Returns
        -------
        set
            A set of events that can be selected.
        """
        possible_events = set()
        for statement in statements:
            if 'request' in statement:  # should be eligible for sets
                if isinstance(statement['request'], Iterable):
                    possible_events.update(statement['request'])
                elif isinstance(statement['request'], BEvent):
                    possible_events.add(statement['request'])
                else:
                    raise TypeError("request parameter should be BEvent or iterable")
        for statement in statements:
            if 'block' in statement:
                if isinstance(statement.get('block'), BEvent):
                    possible_events.discard(statement.get('block'))
                else:
                    possible_events = {x for x in possible_events if x not in statement.get('block')}
        return possible_events

    def select(self, statements, external_events_queue=[]):
        """
        Selects the next event from the given statements and external events queue.

        This method selects the next event uniformly at random from the set of selectable events If no events can be
        selected from the statements, an event from the external events queue will be selected (or `None` is returned
        if the queue is empty).

        Parameters
        ----------
        statements : list
            A list of bthreads sync statements from which an event will be selected.
        external_events_queue : list, optional
            A list of external events that may be selected.

        Returns
        -------
        :class:`BEvent <bppy.model.b_event.BEvent>` or `None`
            The selected event, or `None` if no event can be selected.
        """
        selectable_events = self.selectable_events(statements)
        if selectable_events:
            return random.choice(tuple(selectable_events))
        else:
            if len(external_events_queue) > 0:
                return external_events_queue.pop(0)
            else:
                return None




import unittest
from bppy import *


class TestExamples(unittest.TestCase):

    def test_hello_world(self):
        @b_thread
        def hello():
            yield {request: BEvent("Hello")}

        @b_thread
        def world():
            yield {request: BEvent("World")}
        b_program = BProgram(bthreads=[hello(), world()],
                             event_selection_strategy=SimpleEventSelectionStrategy())
        b_program.run()
        assert True

    def test_hot_cold_all(self):
        @b_thread
        def add_hot():
            yield {request: BEvent("HOT")}
            yield {request: BEvent("HOT")}
            yield {request: BEvent("HOT")}

        @b_thread
        def add_cold():
            yield {request: BEvent("COLD")}
            yield {request: BEvent("COLD")}
            yield {request: BEvent("COLD")}

        @b_thread
        def control_temp():
            e = BEvent("Dummy")
            while True:
                e = yield {waitFor: All(), block: e}

        b_program = BProgram(bthreads=[add_hot(), add_cold(), control_temp()],
                             event_selection_strategy=SimpleEventSelectionStrategy())
        b_program.run()
        assert True

    def test_hot_cold_bath(self):
        @b_thread
        def add_hot():
            yield {request: BEvent("HOT")}
            yield {request: BEvent("HOT")}
            yield {request: BEvent("HOT")}

        @b_thread
        def add_cold():
            yield {request: BEvent("COLD")}
            yield {request: BEvent("COLD")}
            yield {request: BEvent("COLD")}

        @b_thread
        def control_temp():
            while True:
                yield {waitFor: BEvent("COLD"), block: BEvent("HOT")}
                yield {waitFor: BEvent("HOT"), block: BEvent("COLD")}

        b_program = BProgram(bthreads=[add_hot(), add_cold(), control_temp()],
                             event_selection_strategy=SimpleEventSelectionStrategy())
        b_program.run()
        assert True

    def test_hot_cold_dynamic(self):
        @b_thread
        def add_hot():
            yield {request: BEvent("HOT")}
            yield {request: BEvent("HOT")}
            yield {request: BEvent("HOT")}

        @b_thread
        def add_cold():
            yield {request: BEvent("COLD")}
            yield {request: BEvent("COLD")}
            yield {request: BEvent("COLD")}

        @b_thread
        def control_temp(block_event):
            block_event = yield {waitFor: All(), block: block_event}
            b_program.add_bthread(control_temp(block_event))

        b_program = BProgram(bthreads=[add_hot(), add_cold(), control_temp(BEvent("HOT"))],
                             event_selection_strategy=SimpleEventSelectionStrategy())
        b_program.run()
        assert True

    def test_hot_cold_smt(self):
        hot = Bool('hot')
        cold = Bool('cold')

        @b_thread
        def three_hot():
            for i in range(3):
                yield {request: hot, waitFor: hot}

        @b_thread
        def three_cold():
            for j in range(3):
                yield {request: cold, waitFor: cold}

        @b_thread
        def control_temp():
            m = yield {}
            while True:
                if is_true(m[cold]):
                    m = yield {block: cold}
                if is_true(m[hot]):
                    m = yield {block: hot}

        @b_thread
        def mutual_exclusion():
            yield {block: And(cold, hot), waitFor: false}

        b_program = BProgram(bthreads=[three_cold(), three_hot(), mutual_exclusion(), control_temp()],
                             event_selection_strategy=SMTEventSelectionStrategy())
        b_program.run()
        assert True

    def test_external_events(self):
        class External(BEvent):
            pass

        any_external = EventSet(lambda event: isinstance(event, External))

        @b_thread
        def add_external():
            b_program.enqueue_external_event(External("A"))
            b_program.enqueue_external_event(External("B"))
            b_program.enqueue_external_event(External("C"))
            while True:
                yield {waitFor: All()}

        @b_thread
        def act_on_external():
            while True:
                # triggers external events if exists, else terminates the bprogram
                event = yield {block: All(), waitFor: any_external}
                yield {request: BEvent(event.name)}

        b_program = BProgram(bthreads=[add_external(), act_on_external()],
                             event_selection_strategy=SimpleEventSelectionStrategy())
        b_program.run()
        assert True

    def test_tic_tac_toe_priorities(self):
        x = lambda row, col: BEvent('X' + str(row) + str(col))
        o = lambda row, col: BEvent('O' + str(row) + str(col))

        LINES = [[(i, j) for j in range(3)] for i in range(3)] + [[(i, j) for i in range(3)] for j in range(3)] + [
            [(i, i) for i in range(3)]] + [[(i, 3 - i - 1) for i in range(3)]]
        x_lines = [[x(i, j) for (i, j) in line] for line in LINES]
        o_lines = [[o(i, j) for (i, j) in line] for line in LINES]

        any_x = [x(i, j) for i in range(3) for j in range(3)]
        any_o = [o(i, j) for i in range(3) for j in range(3)]
        move_events = EventSet(lambda e: e.name.startswith('X') or e.name.startswith('O'))

        static_event = {
            'OWin': BEvent('OWin'),
            'XWin': BEvent('XWin'),
            'draw': BEvent('Draw')
        }

        @b_thread
        def square_taken(row, col):
            yield {waitFor: [x(row, col), o(row, col)]}
            yield {block: [x(row, col), o(row, col)]}

        @b_thread
        def enforce_turns():
            while True:
                yield {waitFor: any_x, block: any_o}
                yield {waitFor: any_o, block: any_x}

        @b_thread
        def end_of_game():
            yield {waitFor: list(static_event.values())}
            yield {block: All()}

        @b_thread
        def detect_draw():
            for r in range(3):
                for c in range(3):
                    yield {waitFor: move_events}
            yield {request: static_event['draw'], priority: 90}

        @b_thread
        def detect_x_win(line):
            for i in range(3):
                yield {waitFor: line}
            yield {request: static_event['XWin'], priority: 100}

        @b_thread
        def detect_o_win(line):
            for i in range(3):
                yield {waitFor: line}
            yield {request: static_event['OWin'], priority: 100}

        # Preference to put O on the center
        @b_thread
        def center_preference():
            while True:
                yield {request: o(1, 1), priority: 35}

        # Preference to put O on the corners
        @b_thread
        def corner_preference():
            while True:
                yield {request: [o(0, 0), o(0, 2), o(2, 0), o(2, 2)], priority: 20}

        # Preference to put O on the sides
        @b_thread
        def side_preference():
            while True:
                yield {request: [o(0, 1), o(1, 0), o(1, 2), o(2, 1)], priority: 10}

        # player O strategy to add a third O to win
        @b_thread
        def add_third_o(line):
            for i in range(2):
                yield {waitFor: line}
            yield {request: line, priority: 50}

        # player O strategy to prevent a third X
        @b_thread
        def prevent_third_x(xline, oline):
            for i in range(2):
                yield {waitFor: xline}
            yield {request: oline, priority: 40}

        @b_thread
        def block_fork(xfork, ofork):
            for i in range(2):
                yield {waitFor: xfork}
            yield {request: ofork, priority: 30}

        forks22 = [[x(1, 2), x(2, 0)], [x(2, 1), x(0, 2)], [x(1, 2), x(2, 1)]], [o(2, 2), o(0, 2), o(2, 0)]
        forks02 = [[x(1, 2), x(0, 0)], [x(0, 1), x(2, 2)], [x(1, 2), x(0, 1)]], [o(0, 2), o(0, 0), o(2, 2)]
        forks20 = [[x(1, 0), x(2, 2)], [x(2, 1), x(0, 0)], [x(2, 1), x(1, 0)]], [o(2, 0), o(0, 0), o(2, 2)]
        forks00 = [[x(0, 1), x(2, 0)], [x(1, 0), x(0, 2)], [x(0, 1), x(1, 0)]], [o(0, 0), o(0, 2), o(2, 0)]
        forks_diag = [[x(0, 2), x(2, 0)], [x(0, 0), x(2, 2)]], [o(0, 1), o(1, 0), o(1, 2), o(2, 1)]

        # simulate player X
        @b_thread
        def player_x():
            while True:
                yield {request: any_x}

        bprog = BProgram(
            bthreads=[square_taken(i, j) for i in range(3) for j in range(3)] +
                     [enforce_turns(), end_of_game(), detect_draw()] +
                     [detect_x_win(line) for line in x_lines] +
                     [detect_o_win(line) for line in o_lines] +
                     [center_preference(), corner_preference(), side_preference()] +
                     [add_third_o(line) for line in o_lines] +
                     [prevent_third_x(xline, oline) for (xline, oline) in zip(x_lines, o_lines)] +
                     [block_fork(xfork, forks22[1]) for xfork in forks22[0]] +
                     [block_fork(xfork, forks02[1]) for xfork in forks02[0]] +
                     [block_fork(xfork, forks20[1]) for xfork in forks20[0]] +
                     [block_fork(xfork, forks00[1]) for xfork in forks00[0]] +
                     [block_fork(xfork, forks_diag[1]) for xfork in forks_diag[0]] +
                     [player_x()],
            event_selection_strategy=PriorityBasedEventSelectionStrategy(default_priority=0))
        bprog.run()
        assert True


from os import listdir
from os.path import dirname, basename, isfile, join, isdir

# TODO: find a smarter way to import al submodules
from bppy.execution.listeners.b_program_runner_listener import *
from bppy.execution.listeners.print_b_program_runner_listener import *
from bppy.model.event_selection.event_selection_strategy import *
from bppy.model.event_selection.simple_event_selection_strategy import *
from bppy.model.event_selection.solver_based_event_selection_strategy import *
from bppy.model.event_selection.smt_event_selection_strategy import *
from bppy.model.event_selection.experimental_smt_event_selection_strategy import *
from bppy.model.event_selection.priority_based_event_selection_strategy import *
from bppy.model.b_event import *
from bppy.model.bprogram import *
from bppy.model.event_set import *
from bppy.model.sync_statement import *
from bppy.model.b_thread import *
from bppy.utils.z3helper import *




from bppy.model.event_selection.solver_based_event_selection_strategy import SolverBasedEventSelectionStrategy
from bppy.utils.z3helper import *


class SMTEventSelectionStrategy(SolverBasedEventSelectionStrategy):
    """
    A solver-based implementation of EventSelectionStrategy using z3 SMT solver. The strategy used is the one
    presented in the paper `Executing Scenario-Based Specification with Dynamic Generation of Rich Events
    <https://www.wisdom.weizmann.ac.il/~dharel/papers/CCIS2019RichEvents.pdf>`_.
    """

    def is_satisfied(self, event, statement):
        return is_true(event.eval(statement.get('waitFor', true)))  

    # TODO: implement a way to set additional_statement
    def select(self, statements, additional_statement=None):
        if isinstance(additional_statement, list) and len(additional_statement) > 0:
            raise NotImplementedError("SMTEventSelectionStrategy does not support external events.")
        (request, block) = (false, false)

        # Collect request and block statements
        for l in statements:
            request = Or(request, l.get('request', false))
            block = Or(block, l.get('block', false))

        if additional_statement:
            request = Or(request, additional_statement.get('request', false))
            block = Or(block, additional_statement.get('block', false))


        # Compute a satisfying assignment
        sl = Solver()
        sl.add(And(request, Not(block)))
        if sl.check() == sat:
            return sl.model()
        else:
            return None





import unittest
from bppy import *


class TestBProgram(unittest.TestCase):

    def test_advance_bthreads(self):
        @b_thread
        def hello():
            yield {request: BEvent("Hello")}

        @b_thread
        def world():
            yield {request: BEvent("World")}

        b_program = BProgram(bthreads=[hello(), world()],
                             event_selection_strategy=SimpleEventSelectionStrategy())

        b_program.setup()
        b_program.advance_bthreads(b_program.tickets, BEvent("Hello"))
        b_program.advance_bthreads(b_program.tickets, BEvent("World"))
        assert all([len(x) == 0 for x in b_program.tickets]) and len(b_program.tickets) == 2



from z3 import *

true = BoolSort().cast(True)
false = BoolSort().cast(False)


def toFloat(r):
    return float(r.numerator_as_long()) / float(r.denominator_as_long())


def visitor(e, seen):
    if e in seen:
        return
    seen[e] = True
    yield e
    if is_app(e):
        for ch in e.children():
            for e in visitor(ch, seen):
                yield e
        return
    if is_quantifier(e):
        for e in visitor(e.body(), seen):
            yield e
        return


def printVars(fml):
    seen = {}
    for e in visitor(fml, seen):
        if is_const(e) and e.decl().kind() == Z3_OP_UNINTERPRETED:
            print("Variable", e)
        else:
            print(e)


def getVariables(fml):
    seen = {}
    for e in visitor(fml, seen):
        if is_const(e) and e.decl().kind() == Z3_OP_UNINTERPRETED:
            yield e



from bppy import *


@b_thread
def add_hot():
    yield {request: BEvent("HOT")}
    yield {request: BEvent("HOT")}
    yield {request: BEvent("HOT")}


@b_thread
def add_cold():
    yield {request: BEvent("COLD")}
    yield {request: BEvent("COLD")}
    yield {request: BEvent("COLD")}


@b_thread
def control_temp(block_event):
    block_event = yield {waitFor: All(), block: block_event}
    b_program.add_bthread(control_temp(block_event))


if __name__ == "__main__":
    b_program = BProgram(bthreads=[add_hot(), add_cold(), control_temp(BEvent("HOT"))],
                         event_selection_strategy=SimpleEventSelectionStrategy(),
                         listener=PrintBProgramRunnerListener())
    b_program.run()


class BEvent:
    """
    A class to represent a Behavioral Event (BEvent) object.

    Attributes
    ----------
    name : str
        The name of the event.
    data : dict
        Additional data associated with the event.
    """
    def __init__(self, name="", data={}):
        """
        Constructs all the necessary attributes for the BEvent object.

        Parameters
        ----------
        name : str
            The name of the event.
        data : dict
            Additional data associated with the event.
        """
        self.name = name
        self.data = data

    def __key(self):
        return tuple([self.name]) + tuple(str(self.data.items()))

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, BEvent) and self.__key() == other.__key()

    def __repr__(self):
        return "{}(name={},data={})".format(self.__class__.__name__, self.name, self.data)

    def __str__(self):
        return self.__repr__()






from abc import ABC, abstractmethod


class EventSelectionStrategy(ABC):
    """
    A base class used to represent an Event Selection Strategy. This is an abstract class
    that requires the implementation of `select` and `is_satisfied` methods.
    """
    @abstractmethod
    def select(self, statements, external_events_queue=[]):
        """
        Abstract method to select and return the bprogram's next event.

        This method needs to be implemented in child classes. It takes a list of statements
        and an external events queue, and should return the next event to be selected based
        on the strategy being implemented.

        Parameters
        ----------
        statements : list
            A list of bthread statements from which an event will be selected.
        external_events_queue : list, optional
            A list of external events that may be selected.

        Returns
        -------
        Abstract
            Returns the next event to be selected based on the strategy being implemented,
            or `None` if no event can be selected
        """
        pass

    @abstractmethod
    def is_satisfied(self, event, statement):
        """
        Abstract method to check whether a given event satisfies the given sync statement, and the bthread should
        advance.

        This method needs to be implemented in child classes. It takes an event and a
        statement, and should return a boolean indicating whether the event satisfies
        the statement and the bthread should advance according to the strategy being implemented.

        Parameters
        ----------
        event : :class:`BEvent <bppy.model.b_event.BEvent>`
            The event to be checked against the statement.
        statement : dict
            The statement to check the event against.

        Returns
        -------
        Abstract
            Returns a boolean indicating whether the event satisfies the statement
            and the bthread should advance according to the strategy being implemented.
        """
        pass



import setuptools
import os
__location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))
with open(os.path.join(__location__, "README.md"), "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="bppy",
    version="0.0.6",
    author="Tom Yaacov",
    author_email="tomyaacov1210@gmail.com",
    description="BPpy: Behavioral Programming In Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/bThink-BGU/BPpy",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "z3-solver",
    ],
    #python_requires='>=3.6'
)


from bppy import *

class External(BEvent):
    pass


any_external = EventSet(lambda event: isinstance(event, External))

@b_thread
def add_external():
    b_program.enqueue_external_event(External("A"))
    b_program.enqueue_external_event(External("B"))
    b_program.enqueue_external_event(External("C"))
    while True:
        yield {waitFor: All()}

@b_thread
def act_on_external():
    while True:
        # triggers external events if exists, else terminates the bprogram
        event = yield {block: All(), waitFor: any_external}
        yield {request: BEvent(event.name)}


if __name__ == "__main__":
    b_program = BProgram(bthreads=[add_external(), act_on_external()],
                         event_selection_strategy=SimpleEventSelectionStrategy(),
                         listener=PrintBProgramRunnerListener())
    b_program.run()
